unit ADKPathFinder;

{
 PathFinder (A*) pour le projet ADK-ISO (c)2002 Paul TOTH <tothpaul@free.fr>

 http://www.web-synergy.net/naug-land/

}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

uses
 Types;

type
// Interface ncessaire au PathFinder
 IMap=Interface
  function Width :integer;
  function Height:integer;
  function CanWalk(const Pos:TPoint):boolean;
 end;

// La classe qui cherche le chemin entre deux points d'une carte
 TADKPathFinder=class
 private
  fmap  :IMap;
  Step :array of array of integer;
  Dist :array of array of integer;
  Moves:array of TPoint;
  MoveCount:integer;
  CurPos  :TPoint;
  EndPoint:TPoint;
  Path    :array of integer;
  procedure Explore(const Pos:TPoint);
  function BestMove:TPoint;
  function DirectPath(Pos:TPoint):boolean;
  function OneStep(const Pos:TPoint):TPoint;
 public
  constructor Create(AMap:IMap);
  function FindPath(const ASource,ATarget:TPoint):boolean;
  function NextPoint:TPoint;
  function EndOfPath:boolean;
  function PathLen:integer;
  function PathDir(Index:integer):integer;
  property Map: IMap read fMap write fMap;
 end;

implementation

// Dplacement en X et Y suivant une des 8 directions possibles
const
 Delta:array[0..7,0..1] of integer=(
  (+0,-1),(+1,-1),(+1,+0),(+1,+1),
  (+0,+1),(-1,+1),(-1,+0),(-1,-1)
 );

// Renvoie la distance entre deux points
function Distance(const A,B:TPoint):integer;
var
 dx,dy:integer;
begin
 dx:=A.x-B.x;
 dy:=A.y-B.y;
 result:=(dx*dx)+(dy*dy);
end;

constructor TADKPathFinder.Create(AMap:IMap);
begin
 fmap:=AMap;
end;

// Explore les points immdiatement autour (calcul des distances)
procedure TADKPathFinder.Explore(const Pos:TPoint);
var
 stp:integer;
 dir:integer;
 Test:TPoint;
begin
 stp:=Step[Pos.x,Pos.y]+1;
 for dir:=0 to 7 do begin
  Test.x:=Pos.x+Delta[dir,0]; if (Test.x<0)or(Test.x>=fmap.Width) then continue;
  Test.y:=Pos.y+Delta[dir,1]; if (Test.y<0)or(Test.y>=fmap.Height) then continue;
  if (Step[Test.x,Test.y]=0) and fmap.CanWalk(Test) then begin
   Step[Test.x,Test.y]:=stp;
   Moves[MoveCount]:=Test;
   inc(MoveCount);
   Dist[Test.x,Test.y]:=Distance(Test,EndPoint);
  end;
 end;
end;

// Retourne le point le plus proche de la cible parmis tous les points explors
function TADKPathFinder.BestMove:TPoint;
var
 i,best:integer;
 Dst:integer;
 Index:integer;
begin
 Index:=0;
 with Moves[0] do best:=Dist[x,y];
 for i:=1 to MoveCount-1 do begin
  with Moves[i] do Dst:=Dist[x,y];
  if Dst<best then begin
   Index:=i;
   best:=Dst;
  end;
 end;
 Result:=Moves[Index];
 dec(MoveCount);
 Moves[Index]:=Moves[MoveCount];
end;

// Recherche un chemin entre deux points
function TADKPathFinder.FindPath(const ASource,ATarget:TPoint):boolean;
var
 Pos:TPoint;
begin
 Result:=false;
 if (ASource.X<0)or(ASource.X>=fmap.Width)
  or(ASource.Y<0)or(ASource.Y>=fmap.Height)
  or(ATarget.X<0)or(ATarget.X>=fmap.Width)
  or(ATarget.Y<0)or(ATarget.Y>=fmap.Height) then exit;
 CurPos  :=ASource;
 EndPoint:=ATarget;
 Step:=nil; // pour tre certain d'avoir des valeurs  0
 SetLength(Step,fmap.Width,fmap.Height);
 SetLength(Dist,fmap.Width,fmap.Height);
 Setlength(Moves,fmap.Width*fmap.Height);

 // 1) trouver un chemin entre les deux points
 Pos:=ASource;
 Step[Pos.x,Pos.y]:=1;
 MoveCount:=0;
 repeat
 // explorer les cases immdiatement autour
  Explore(Pos);
  if MoveCount=0 then exit; // aucun chemin possible
 // trouver la meilleur case
  Pos:=BestMove;
 // jusqu' ce qu'on atteigne la destination
 until (Pos.x=EndPoint.x)and(Pos.y=EndPoint.y);

 // libration partielle de la mmoire
 Moves:=nil;
 Dist:=nil;

 // il existe une chemin entre les deux points
 Result:=True;
end;

// Renvoie le prochain point  atteindre (en ligne droite)
function TADKPathFinder.NextPoint:TPoint;
begin
 Result:=EndPoint;
 if Step=nil then exit;
 while not DirectPath(Result) do Result:=OneStep(Result);
 CurPos:=Result;
// si c'est la dernire tape, on libre "Step"
 if (CurPos.x=EndPoint.x)and(CurPos.y=EndPoint.y) then Step:=nil;
end;

// dtermine si le point "pos" peut atteindre "CurPos" en ligne droite
// on en profite pour calculer "Path"
function TADKPathFinder.DirectPath(Pos:TPoint):boolean;
var
 dx,dy:integer;
 ix,iy:integer;
 ex,ey:integer;
 ee,ii:integer;
   dir:integer;
begin
 dx:=CurPos.x-Pos.x; if dx>0 then ix:=1 else begin dx:=-dx; ix:=-1 end;
 dy:=CurPos.y-Pos.y; if dy>0 then iy:=1 else begin dy:=-dy; iy:=-1 end;
 ex:=dx;
 ey:=dy;
 if ex<ey then ee:=ey else ee:=ex;
 SetLength(Path,ee);
 for ii:=ee-1 downto 0 do begin
  inc(ex,dx);
  inc(ey,dy);
  dir:=0;
  if ex>ee then begin
   dec(ex,ee);
   inc(Pos.x,ix);
   if ix<0 then dir:=3 else dir:=7;
  end;
  if ey>ee then begin
   dec(ey,ee);
   inc(Pos.y,iy);
   case dir of
    0: if iy<0 then dir:=5 else dir:=1;
    3: if iy<0 then dir:=4 else dir:=2;
    7: if iy<0 then dir:=6 else dir:=8;
   end;
  end;
  Path[ii]:=dir;
  if fmap.CanWalk(Pos)=False then begin
   Result:=False;
   exit;
  end;
 end;
 Result:=True;
end;

// dtermine si on a atteind la destination
function TADKPathFinder.EndOfPath:boolean;
begin
 Result:=(Step=nil);
end;

// dplace le point vers la case qui demande le moins d'tapes (step)
function TADKPathFinder.OneStep(const Pos:TPoint):TPoint;
var
 best:integer;
 dir:integer;
 Test:TPoint;
 stp :integer;
begin
 best:=Step[Pos.x,Pos.y];
 for dir:=0 to 7 do begin
  Test.x:=Pos.x+Delta[dir,0]; if (Test.x<0)or(Test.x>=fmap.Width) then continue;
  Test.y:=Pos.y+Delta[dir,1]; if (Test.y<0)or(Test.y>=fmap.Height) then continue;
  stp:=Step[Test.x,Test.y];
  if (stp>0)and(stp<best) then begin
   Result:=Test;
   best:=stp;
  end;
 end;
end;

// longeur du chemin vers le point retourn par NextPoint
function TADKPathFinder.PathLen:integer;
begin
 result:=Length(Path);
end;

// les directions  prendre dans le chemin ci-dessus
// (1=Nord, 2=NordEst, 3=Est...7=Ouest, 8=NordOuest).
function TADKPathFinder.PathDir(Index:integer):integer;
begin
 if Index>=Length(Path) then
  result:=0
 else
  result:=Path[Index];
end;

end.
